//+------------------------------------------------------------------+
//|                                                    TicksInfo.mqh |
//|                        Copyright 2019, MetaQuotes Software Corp. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2019, MetaQuotes Software Corp."
#property link      "https://www.mql5.com"
#property version   "1.00"
#property strict

#define SECONDSPERDAY 86400
#define MAXSECONDSOFGAP 60

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CTicksInfo
{
private:
   string symArr[];
   string symbols;
   int files[];
   MqlTick nextTicks[], prevTicks[];
   bool symAborted[];

   int getSymIdx(string sym);
   void refreshTicks(datetime dt) { refreshTicks(long(dt)); }
   void refreshTicks(long msc);

public:
   CTicksInfo();
   ~CTicksInfo();
   
   bool init(string &arr[]);
   void deInit();
   bool getTick(string sym, MqlTick &tick);
};
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CTicksInfo::CTicksInfo()
{
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CTicksInfo::~CTicksInfo()
{
}
//+------------------------------------------------------------------+

bool CTicksInfo::init(string &arr[])
{
   if (!IsTesting())
      return (true);

   int newSize = ArrayResize(symArr,ArraySize(arr));
   if (newSize!=ArraySize(arr))
   {
      Print("CTicksInfo::init - Failed to resize array!");
      return (false);
   }
   
   int copiedItems = ArrayCopy(symArr,arr);
   if (copiedItems!=newSize)
   {
      Print("CTicksInfo::init - Failed to copy array!");
      return (false);
   }

   ArrayResize(files,newSize);
   ArrayResize(nextTicks,newSize);
   ArrayResize(prevTicks,newSize);
   ArrayResize(symAborted,newSize);
   
   for (int i=0; i<newSize; i++)
   {
      symbols = symbols+symArr[i];
      files[i] = FileOpen(symArr[i],FILE_READ|FILE_BIN);
      if (files[i]==INVALID_HANDLE)
      {
         Print("CTicksInfo::init - Failed to open file ", symArr[i], ". Error code: ", _LastError);
         symAborted[i] = true;
      }
      else
      {
         symAborted[i] = false;
         FileSeek(files[i],0,SEEK_SET);
      }
   }
   
   return (true);
}

void CTicksInfo::deInit(void)
{
   if (!IsTesting())
      return;

   for (int i=0; i<ArraySize(symArr); i++)
      FileClose(files[i]);
}

int CTicksInfo::getSymIdx(string sym)
{
   int pos = StringFind(symbols,sym);
   if (pos<0)
      return (pos);
   else
      return (pos/6);
}

void CTicksInfo::refreshTicks(long msc)
{
   if (!IsTesting())
      return;
   
   for (int i=0; i<ArraySize(symArr); i++)
   {
      if (files[i]>0 && !symAborted[i])
      {
         while (msc>nextTicks[i].time_msc || nextTicks[i].time_msc==0)
         {
            prevTicks[i] = nextTicks[i];
            uint bytesRead = FileReadStruct(files[i],nextTicks[i]);
            if (bytesRead!=sizeof(nextTicks[i]))
            {
               Print ("CTicksInfo::refreshTicks - Error reading tick data for ", symArr[i], ". Error code: ", _LastError);
               symAborted[i] = true;
               break;
            }
         }
      }
   }
}

bool CTicksInfo::getTick(string sym, MqlTick &tick)
{
   if (!IsTesting())
      return (SymbolInfoTick(sym,tick));
   
   static datetime lastTick = 0;
   if (lastTick!=0)
   {
      long timeDiff = TimeCurrent()-lastTick;
      if (timeDiff>MAXSECONDSOFGAP && timeDiff<SECONDSPERDAY)
         Print ("CTicksInfo::getTick - Potential data gap detected between ", datetime(lastTick), " and ", TimeCurrent(), " for ", sym);
   }
   
   int symIdx = getSymIdx(sym);
   if (symIdx<0 || symAborted[symIdx])
   {
      bool result = SymbolInfoTick(sym,tick);
      refreshTicks(tick.time_msc);
      return (result);
   }
   else
   {
      if (prevTicks[symIdx].ask==0)
         refreshTicks(TimeCurrent()*1000);
      tick = prevTicks[symIdx];
   }
   
   lastTick = TimeCurrent();
   
   return (true);
}
